import ipaddress
import unittest

from ndk import fields


class FieldTestCase(unittest.TestCase):

    def test_Field_is_works(self):
        f = fields.Field()
        assert f.primary_key == False
        assert f.required == False
        assert f.composite_key == False
        assert f.default == None

    def test_normalize_name_is_works(self):
        assert 'foo-bar' == fields.Field.normalize_name('Foo BAR')

    def test_type_conversion_of_default(self):
        str_default = fields.Field(default='foo')
        assert isinstance(str_default.default, str)

        int_defualt = fields.Field(default=1234)
        assert isinstance(int_defualt.default, int)


class StringFieldTestCase(unittest.TestCase):

    def test_StringField_is_works(self):
        f = fields.StringField()
        assert f.primary_key == False
        assert f.required == False
        assert f.composite_key == False
        assert f.default == None

    def test_type_conversion_of_default_in_StringField(self):
        str_default = fields.StringField(default='foo')
        assert isinstance(str_default.default, str)

        int_default = fields.StringField(default=1234)
        assert isinstance(int_default.default, str)
        assert '1234' == int_default.default


class IntegerFieldTestCase(unittest.TestCase):

    def test_IntegerField_is_works(self):
        f = fields.IntegerField()
        assert f.primary_key == False
        assert f.required == False
        assert f.composite_key == False
        assert f.default == None

    def test_type_conversion_of_default_in_IntegerField(self):
        with self.assertRaises(ValueError):
            fields.IntegerField(default='foo')

        f = fields.IntegerField(default='1234')
        assert isinstance(f.default, int)
        assert 1234 == f.default

        f = fields.IntegerField(default=1234)
        assert isinstance(f.default, int)
        assert 1234 == f.default


class BooleanFieldTestCase(unittest.TestCase):

    def test_BooleanField_is_works(self):
        f = fields.BooleanField()
        assert f.primary_key == False
        assert f.required == False
        assert f.composite_key == False
        assert f.default == None

    def test_type_conversion_of_default_in_BooleanField(self):
        # Nagios Object use [0/1] to represent the Boolean value
        f = fields.BooleanField(default=True)
        assert isinstance(f.default, int)

        f = fields.BooleanField(default=True)
        assert f.default

        f = fields.BooleanField(default=0)
        assert not f.default



import unittest

from ndk import core
from ndk import exceptions


class StackTestCase(unittest.TestCase):
    """A Stack TestCase"""

    def test_to_create_stack(self):
        stack = core.Stack('StackTesting')
        assert stack.name == 'StackTesting'

    def test_to_push_object_to_the_stack(self):
        stack = core.Stack('StackTesting')
        # Make a fake Nagios Object
        obj = type('foo', (object,), dict(
            __object_type__='host', pk='bar', name='baz'))
        stack.push(obj)
        assert 'host' in stack.objects
        assert obj == stack.objects['host']['bar']

    def test_to_raise_DuplicateError_if_pk_already_exist(self):
        stack = core.Stack('StackTesting')
        # Make a fake Nagios Object
        obj = type('foo', (object,), dict(
            __object_type__='host', pk='bar', name='baz'))
        stack.push(obj)
        with self.assertRaises(exceptions.DuplicateError):
            #  Make a new Object with a same type and pk
            obj = type('Fo', (object,), dict(
                __object_type__='host', pk='bar', name='BA'))
            stack.push(obj)



from ndk import core, fields


class ContactGroupConstruct(core.Object):
    """
    L1 Construct: Nagios::Object::ContactGroup

    This construct correspond directly to object defined by Nagios.
    """

    class Meta:
        object_type = 'contactgroup'

    contactgroup_name = fields.StringField(primary_key=True, required=True)
    alias = fields.StringField(required=True)
    members = fields.ForeignKey(relation='Contact')
    contactgroup_members = fields.ForeignKey(relation='ContactGroup')

    def __init__(
            self, stack, contactgroup_name, alias, members=None,
            contactgroup_members=None):
        super().__init__(
            stack=stack, contactgroup_name=contactgroup_name,
            alias=alias, members=members,
            contactgroup_members=contactgroup_members)


class ContactGroup(ContactGroupConstruct):
    """
    L2 Construct: Nagios::Object::ContactGroup

    L2 encapsulate L1 modules, it is developed to address specific use
    cases and sensible defaults.
    """

    def __init__(self, stack, contactgroup_name, alias=None, **kwargs):
        alias = alias or contactgroup_name
        super().__init__(stack, contactgroup_name, alias, **kwargs)



import unittest

from ndk import core, objects
from ndk.objects import ServiceGroup, ServiceGroupConstruct


class ServiceGroupTestCase(unittest.TestCase):

    def test_L1_construct_is_works(self):
        stack = core.Stack('ServiceGroupTesting')
        cg = ServiceGroupConstruct(
            stack, servicegroup_name='foo', alias='bar', members='baz')
        assert cg.servicegroup_name == 'foo'

    def test_L2_construct_is_works(self):
        stack = core.Stack('ServiceGroupTesting')
        cg = ServiceGroup(stack, servicegroup_name='foo',
                          alias='bar', members=['bar', 'baz'])
        assert cg.members == ['bar', 'baz']



from enum import Enum

from ndk import core, fields


class ChoiceMixin:

    @classmethod
    def choices(cls):
        for name, member in cls.__members__.items():
            if not name == 'NO':
                yield member

    @classmethod
    def all(cls):
        return list(cls.choices())

    @classmethod
    def empty(cls):
        return cls.NO


class InitialState(ChoiceMixin, Enum):
    """
    Override the initial state for a Host

    By default Nagios will assume that all services are in OK states when it starts.
    You can override the initial state for a host by using this directive.

    Valid options:
      - o = OK
      - w = WARNING
      - u = UNKNOWN
      - c = CRITICAL
    """

    OK = 'o'
    WARNING = 'w'
    UNKNOWN = 'u'
    CRITICAL = 'c'


class FlapDetection(ChoiceMixin, Enum):
    """
    This directive is used to determine what service states the flap detection
    logic will use for this service.

    Valid options:
      - o = OK
      - w = WARNING
      - u = UNKNOWN
      - c = CRITICAL
    """

    OK = 'o'
    WARNING = 'w'
    UNKNOWN = 'u'
    CRITICAL = 'c'


class Notification(ChoiceMixin, Enum):
    """
    This directive is used to determine
    when notifications for the service should be sent out.

    If you do not specify any notification options,
    Nagios will assume that you want notifications to be sent out
    for all possible states.

    Valid options:
      - w = send notifications on a WARNING state
      - c = send notifications on a CRITICAL state
      - u = send notifications on an UNKNOWN state
      - r = send notifications on recoveries (OK state)
      - f = send notifications when the service starts and stops flapping
      - s = send notifications when scheduled downtime starts and ends
      - n = no service notifications will be sent out
    """

    WARNING = 'w'
    CRITICAL = 'c'
    UNKNOWN = 'u'
    RECOVERY = 'r'
    FLAPPING = 'f'
    SCHEDULED = 's'
    NO = 'n'


class Stalking(ChoiceMixin, Enum):
    """
    This directive determines which service states "stalking" is enabled for.

    Note: As of Core 4.4.0 you can use the N option to log event states when notifications are sent out.

    Valid options:
      - o = stalk on OK states
      - w = stalk on WARNING states
      - u = stalk on UNKNOWN states
      - c = stalk on CRITICAL states
    """

    OK = 'o'
    WARNING = 'w'
    UNKNOWN = 'u'
    CRITICAL = 'c'


class ServiceConstruct(core.Object):
    """
    L1 Construct: Nagios::Object::Service
    """

    class Meta:
        object_type = 'service'

    host_name = fields.StringField(primary_key=True, required=True)
    hostgroup_name = fields.ForeignKey(relation='HostGroup')
    service_description = fields.StringField(primary_key=True, required=True)
    display_name = fields.StringField()
    parents = fields.ForeignKey(relation='Service')
    importance = fields.IntegerField()
    servicegroups = fields.ForeignKey(relation='ServiceGroup')
    is_volatile = fields.BooleanField()
    check_command = fields.StringField(required=True)
    initial_state = fields.ChoiceField(choices=InitialState)
    max_check_attempts = fields.IntegerField(required=True)
    check_interval = fields.IntegerField(required=True)
    retry_interval = fields.IntegerField(required=True)
    active_checks_enabled = fields.BooleanField()
    passive_checks_enabled = fields.BooleanField()
    check_period = fields.ForeignKey(relation='TimePeriod', required=True)
    obsess_over_service = fields.BooleanField()
    check_freshness = fields.BooleanField()
    freshness_threshold = fields.IntegerField()
    event_handler = fields.ForeignKey(relation='Command')
    event_handler_enabled = fields.BooleanField()
    low_flap_threshold = fields.IntegerField()
    high_flap_threshold = fields.IntegerField()
    flap_detection_enabled = fields.BooleanField()
    flap_detection_options = fields.ChoiceField(
        choices=FlapDetection)
    process_perf_data = fields.BooleanField()
    retain_status_information = fields.BooleanField()
    retain_nonstatus_information = fields.BooleanField()
    notification_interval = fields.IntegerField(required=True)
    first_notification_delay = fields.IntegerField()
    notification_period = fields.ForeignKey(
        relation='TimePeriod', required=True)
    notification_options = fields.ChoiceField(
        choices=Notification)
    notifications_enabled = fields.BooleanField()
    contacts = fields.ForeignKey(relation='Contact', composite_key=True)
    contact_groups = fields.ForeignKey(
        relation='ContactGroup', composite_key=True)
    stalking_options = fields.ChoiceField(choices=Stalking)
    notes = fields.StringField()
    notes_url = fields.StringField()
    action_url = fields.StringField()
    icon_image = fields.StringField()
    icon_image_alt = fields.StringField()

    def __init__(
            self, stack, host_name, service_description, check_command,
            check_interval, retry_interval, check_period,
            notification_interval, notification_period, hostgroup_name=None,
            display_name=None, parents=None, importance=None,
            servicegroups=None, is_volatile=None, initial_state=None,
            max_check_attempts=None, active_checks_enabled=None,
            passive_checks_enabled=None, obsess_over_service=None,
            check_freshness=None, freshness_threshold=None, event_handler=None,
            event_handler_enabled=None, low_flap_threshold=None,
            high_flap_threshold=None, flap_detection_enabled=None,
            flap_detection_options=None, process_perf_data=None,
            retain_status_information=None, retain_nonstatus_information=None,
            first_notification_delay=None, notification_options=None,
            notifications_enabled=None, contacts=None, contact_groups=None,
            stalking_options=None, notes=None, notes_url=None, action_url=None,
            icon_image=None, icon_image_alt=None):
        super().__init__(
            stack, host_name=host_name, hostgroup_name=hostgroup_name,
            service_description=service_description, display_name=display_name,
            parents=parents, importance=importance,
            servicegroups=servicegroups, is_volatile=is_volatile,
            check_command=check_command, initial_state=initial_state,
            max_check_attempts=max_check_attempts,
            check_interval=check_interval, retry_interval=retry_interval,
            active_checks_enabled=active_checks_enabled,
            passive_checks_enabled=passive_checks_enabled,
            check_period=check_period, obsess_over_service=obsess_over_service,
            check_freshness=check_freshness,
            freshness_threshold=freshness_threshold,
            event_handler=event_handler,
            event_handler_enabled=event_handler_enabled,
            low_flap_threshold=low_flap_threshold,
            high_flap_threshold=high_flap_threshold,
            flap_detection_enabled=flap_detection_enabled,
            flap_detection_options=flap_detection_options,
            process_perf_data=process_perf_data,
            retain_status_information=retain_status_information,
            retain_nonstatus_information=retain_nonstatus_information,
            notification_interval=notification_interval,
            first_notification_delay=first_notification_delay,
            notification_period=notification_period,
            notification_options=notification_options,
            notifications_enabled=notifications_enabled, contacts=contacts,
            contact_groups=contact_groups, stalking_options=stalking_options,
            notes=notes, notes_url=notes_url, action_url=action_url,
            icon_image=icon_image, icon_image_alt=icon_image_alt)


class Service(ServiceConstruct):
    """
    L2 Construct: Nagios::Object::Service
    """

    def __init__(
            self, stack, host_name, service_description, check_command,
            notification_period, check_interval=3, retry_interval=2,
            check_period=5, notification_interval=60, max_check_attempts=5,
            active_checks_enabled=True, passive_checks_enabled=False,
            event_handler=None, event_handler_enabled=False,
            flap_detection_enabled=True, flap_detection_options=None,
            first_notification_delay=None, notification_options=None,
            notifications_enabled=True, contacts=None, contact_groups=None, **
            kwargs):
        super().__init__(
            stack, host_name=host_name,
            service_description=service_description,
            check_command=check_command, max_check_attempts=max_check_attempts,
            check_interval=check_interval, retry_interval=retry_interval,
            active_checks_enabled=active_checks_enabled,
            passive_checks_enabled=passive_checks_enabled,
            check_period=check_period, event_handler=event_handler,
            event_handler_enabled=event_handler_enabled,
            flap_detection_enabled=flap_detection_enabled,
            flap_detection_options=flap_detection_options,
            notification_interval=notification_interval,
            first_notification_delay=first_notification_delay,
            notification_period=notification_period,
            notification_options=notification_options,
            notifications_enabled=notifications_enabled, contacts=contacts,
            contact_groups=contact_groups, **kwargs)



from enum import Enum

from ndk import core, fields


class ChoiceMixin:

    @classmethod
    def choices(cls):
        for name, member in cls.__members__.items():
            if not name == 'NO':
                yield member

    @classmethod
    def all(cls):
        return list(cls.choices())

    @classmethod
    def empty(cls):
        return cls.NO


class HostNotifications(ChoiceMixin, Enum):
    """
    This directive is used to define the host states for
    which notifications can be sent out to this contact.

    Valid options:
        - d = notify on DOWN host states
        - u = notify on UNREACHABLE host states
        - r = notify on host recoveries (UP states)
        - f = notify when the host starts and stops flapping
        - s = notify when host scheduled downtime starts and ends
        - n = the contact will not receive any type of host notifications
    """

    DOWN = 'd'
    UNREACHABLE = 'u'
    RECOVERY = 'r'
    FLAPPING = 'f'
    SCHEDULED = 's'
    NO = 'n'


class ServiceNotifications(ChoiceMixin, Enum):
    """
    This directive is used to define the service states for
    which notifications can be sent out to this contact.

    Valid options:
        - w = notify on WARNING service states
        - u = notify on UNKNOWN service states
        - c = notify on CRITICAL service states
        - r = notify on service recoveries (OK states)
        - f = notify when the service starts and stops flapping
        - n = the contact will not receive any type of service notifications.
    """

    WARNING = 'w'
    UNKNOWN = 'u'
    CRITICAL = 'c'
    RECOVERY = 'r'
    FLAPPING = 'f'
    SCHEDULED = 's'
    NO = 'n'


class ContactConstruct(core.Object):
    """
    L1 Construct: Nagios::Object::Contact

    This construct correspond directly to object defined by Nagios.
    """

    class Meta:
        object_type = 'contact'

    contact_name = fields.StringField(primary_key=True, required=True)
    alias = fields.StringField()
    contactgroups = fields.ForeignKey(relation='ContactGroup')
    minimum_importance = fields.IntegerField()
    host_notifications_enabled = fields.BooleanField(required=True)
    service_notifications_enabled = fields.BooleanField(required=True)
    host_notifications_period = fields.ForeignKey(
        relation='TimePeriod', required=True)
    service_notifications_period = fields.ForeignKey(
        relation='TimePeriod', required=True)
    host_notifications_options = fields.ChoiceField(
        choices=HostNotifications, required=True)
    service_notifications_options = fields.ChoiceField(
        choices=ServiceNotifications, required=True)
    host_notification_commands = fields.ForeignKey(
        relation='Command', required=True)
    service_notification_commands = fields.ForeignKey(
        relation='Command', required=True)
    email = fields.StringField()
    pager = fields.StringField()
    addressx = fields.StringField()
    can_submit_commands = fields.BooleanField()
    retain_status_information = fields.BooleanField()
    retain_nonstatus_information = fields.BooleanField()

    def __init__(
            self, stack, contact_name, host_notifications_enabled,
            service_notifications_enabled, host_notifications_period,
            service_notifications_period, host_notifications_options,
            service_notifications_options, host_notification_commands,
            service_notification_commands, alias=None, contactgroups=None,
            minimum_importance=None, email=None, pager=None, addressx=None,
            can_submit_commands=None, retain_status_information=None,
            retain_nonstatus_information=None):
        super().__init__(
            stack, contact_name=contact_name,
            host_notifications_enabled=host_notifications_enabled,
            service_notifications_enabled=service_notifications_enabled,
            host_notifications_period=host_notifications_period,
            service_notifications_period=service_notifications_period,
            host_notifications_options=host_notifications_options,
            service_notifications_options=service_notifications_options,
            host_notification_commands=host_notification_commands,
            service_notification_commands=service_notification_commands,
            alias=alias, contactgroups=contactgroups,
            minimum_importance=minimum_importance, email=email, pager=pager,
            addressx=addressx, can_submit_commands=can_submit_commands,
            retain_status_information=retain_status_information,
            retain_nonstatus_information=retain_nonstatus_information)


class Contact(ContactConstruct):
    """
    L2 Construct: Nagios::Object::Contact

    L2 encapsulate L1 modules, it is developed to address specific use
    cases and sensible defaults.
    """

    def __init__(
            self, stack, contact_name, host_notifications_period,
            service_notifications_period, host_notification_commands,
            service_notification_commands, host_notifications_enabled=True,
            service_notifications_enabled=True,
            host_notifications_options=HostNotifications.all(),
            service_notifications_options=ServiceNotifications.all(),
            **kwargs):
        super().__init__(
            stack=stack, contact_name=contact_name,
            host_notifications_period=host_notifications_period,
            service_notifications_period=service_notifications_period,
            host_notification_commands=host_notification_commands,
            service_notification_commands=service_notification_commands,
            host_notifications_enabled=host_notifications_enabled,
            service_notifications_enabled=service_notifications_enabled,
            host_notifications_options=host_notifications_options,
            service_notifications_options=service_notifications_options,
            **kwargs)


class Email(Contact):
    """
    L3 Construct: Nagios::Object::Contact

    L3 declare a resource to create particular use cases.
    """

    def __init__(
            self, stack, contact_name, email, host_notifications_period,
            service_notifications_period, host_notification_commands,
            service_notification_commands, **kwargs):
        super().__init__(
            stack, contact_name=contact_name, email=email,
            host_notifications_period=host_notifications_period,
            service_notifications_period=service_notifications_period,
            host_notification_commands=host_notification_commands,
            service_notification_commands=service_notification_commands,
            **kwargs)



import importlib
from dataclasses import dataclass, field
from enum import Enum
from ipaddress import IPv4Address
from typing import Any, List

from ndk.exceptions import IntegrityError


@dataclass
class Field:
    primary_key: bool = False
    composite_key: bool = False
    required: bool = False
    default: str = None
    field_type: type = None

    def __post_init__(self):
        if self.default is not None and self.field_type is not None:
            self.default = self.serializer(self.default)

    @staticmethod
    def normalize_name(name):
        return str(name).lower().replace(' ', '-')

    def serializer(self, value):
        return self.field_type(value)


@dataclass
class StringField(Field):
    field_type: str = str


@dataclass
class IntegerField(Field):
    field_type: int = int


@dataclass
class BooleanField(Field):
    field_type: bool = bool

    def serializer(self, value):
        """Nagios Object use [0/1] to represent the Boolean value"""
        if self.field_type(value):
            return 1
        return 0


@dataclass
class Ipv4Field(Field):
    field_type: IPv4Address = IPv4Address


@dataclass
class ChoiceField(Field):
    field_type: List[str] = field(default_factory=list)
    choices: Enum = None

    def __post_init__(self):
        if self.primary_key:
            raise IntegrityError('ChoiceField can not be a Primary Key')
        if not self.choices:
            raise IntegrityError('`.choices` is required in ChoiceField')
        if not issubclass(self.choices, Enum):
            raise IntegrityError('`.choices` must be Enum instance')
        super().__post_init__()

    def serializer(self, value):
        if value:
            if not isinstance(value, list):
                value = [value]
            return ','.join((x.value for x in value))
        return value


@dataclass
class ForeignKey(Field):
    field_type: type = str
    relation: str = None

    def __post_init__(self):
        if not self.relation:
            raise IntegrityError('.relation is requried field in ForeignKey')
        super().__post_init__()

    def serializer(self, obj):
        module = importlib.import_module('.objects', package='ndk')
        relation = getattr(module, self.relation)
        if not isinstance(obj, relation):
            raise IntegrityError(
                f'{obj} object must be a {relation} instance.')
        return self.field_type(obj)



import unittest

from ndk import core, objects
from ndk.objects import Contact, ContactConstruct
from ndk.objects.contact import HostNotifications, ServiceNotifications


class ContactTestCase(unittest.TestCase):

    def test_constructs_are_works(self):
        stack = core.Stack('ContactTesting')
        _24x7 = objects.TwentyFourSeven(stack)
        cmd = objects.command.Ping(stack)  # fake cmd
        contact = ContactConstruct(
            stack, contact_name='foo',
            host_notifications_enabled=True,
            service_notifications_enabled=True,
            host_notifications_period=_24x7,
            service_notifications_period=_24x7,
            host_notifications_options=HostNotifications.choices(),
            service_notifications_options=ServiceNotifications.choices(),
            host_notification_commands=cmd,
            service_notification_commands=cmd)
        assert contact.contact_name == 'foo'

        contact = Contact(
            stack, contact_name='bar',
            host_notifications_period=_24x7,
            service_notifications_period=_24x7,
            host_notification_commands=cmd,
            service_notification_commands=cmd)
        assert contact.host_notifications_enabled == True
        assert 'host_notifications_options    d,u,r,f,s' in contact.synth()

    def test_L3_construct_is_works(self):
        stack = core.Stack('ContactTesting')
        _24x7 = objects.TwentyFourSeven(stack)
        cmd = objects.command.Ping(stack)  # fake cmd
        contact = objects.contact.Email(
            stack, contact_name='bar', email='foo@bar.baz',
            host_notifications_period=_24x7,
            service_notifications_period=_24x7,
            host_notification_commands=cmd,
            service_notification_commands=cmd)
        assert contact.host_notifications_enabled == True
        assert 'host_notifications_options    d,u,r,f,s' in contact.synth()
        assert contact.email == 'foo@bar.baz'



import unittest

# from . import models
from ndk import core, exceptions, fields, objects


class NagiosTestCase(unittest.TestCase):

    def test_raise_IntegrityError_when_create_new_class(self):
        with self.assertRaises(exceptions.IntegrityError):
            class EmptyObject(core.Object):
                pass

        with self.assertRaises(exceptions.IntegrityError):
            class ObjectWithoutPK(core.Object):
                class Meta:
                    object_type = 'host'

                directive = fields.Field()

        with self.assertRaises(exceptions.IntegrityError):
            class ObjectWithoutObjectType(core.Object):

                directive = fields.Field(primary_key=True)

    def test_to_create_new_class(self):
        class GoodObject(core.Object):
            class Meta:
                object_type = 'host'
            host_name = fields.Field(primary_key=True)

        assert 'host' == GoodObject.__object_type__
        assert 'host_name' in GoodObject.__mappings__
        assert 'host_name' in GoodObject.__primary_key__
        assert not GoodObject.__composite_key__
        stack = core.Stack('ObjectTesting')
        host = GoodObject(stack, host_name='foo')

    def test_if_pk_has_certain_format(self):
        class ObjectWithOenPK(core.Object):
            class Meta:
                object_type = 'foo'
            host_name = fields.StringField(primary_key=True)

        class ObjectWithTwoPKs(core.Object):
            class Meta:
                object_type = 'foo'
            host_name = fields.StringField(primary_key=True)
            alias = fields.StringField(primary_key=True)

        stack = core.Stack('ObjectTesting')
        host = ObjectWithOenPK(stack, host_name='Foo Bar')
        assert host.pk == 'foo-bar'

        host = ObjectWithTwoPKs(stack, host_name='Foo Bar', alias='BAZ')
        assert host.pk == 'foo-bar::baz'

    def test_required_is_works_in_object(self):
        class Host(core.Object):
            class Meta:
                object_type = 'asdf'

            host_name = fields.StringField(required=True, primary_key=True)
            address = fields.Ipv4Field(required=True)

        stack = core.Stack('FieldTesting')
        with self.assertRaises(exceptions.IntegrityError):
            host = Host(stack, host_name='foo')
            host.is_valid()

        host = Host(stack, host_name='bar', address='127.0.0.1')
        host.is_valid()

    def test_synth_is_works(self):
        stack = core.Stack('ObjectTesting')
        _7x24 = objects.TwentyFourSeven(stack, 'tp-24x7')
        cmd = objects.command.Ping(stack)
        host = objects.Host(
            stack, host_name='foo', address='127.0.0.1',
            check_period=_7x24,
            notification_period=_7x24,
            check_command=cmd)

        define = '\n'.join((
            'define command {',
            '    command_name    check-host-alive',
            '    command_line    $USER1$/check_ping -H $HOSTADDRESS$ -w 30.0,80% -c 50.0,100%',
            '}'
        ))
        assert define == cmd.synth()
        assert 'check_period    tp-24x7' in host.synth()
        assert 'notification_options    d,u,r,f,s' in host.synth()
        assert define in stack.synth()



from ndk import core, fields


class CommandConstruct(core.Object):
    """
    L1 Construct: Nagios::Object::Command

    This construct correspond directly to Command defined by Nagios.
    """

    class Meta:
        object_type = 'command'

    command_name = fields.StringField(primary_key=True, required=True)
    command_line = fields.StringField(required=True)

    def __init__(self, stack, command_name, command_line):
        super().__init__(stack,
                         command_name=command_name,
                         command_line=command_line)


class Command(CommandConstruct):
    """
    L2 Construct: Nagios::Object::Command

    Command encapsulate L1 modules, it is developed to address specific use 
    cases and sensible defaults.
    """

    def __init__(self, stack, **kwargs):
        # All fields are requred in the construct
        super().__init__(stack, **kwargs)


class Ping(Command):
    """
    L3 Construct: Nagios::Object::Command

    L3 declare a resource to create particular use cases.
    """

    def __init__(self, stack, command_name=None):
        name = command_name or 'check-host-alive'
        cmd = '$USER1$/check_ping -H $HOSTADDRESS$ -w 30.0,80% -c 50.0,100%'
        super().__init__(stack, command_name=name, command_line=cmd)



from ndk import core, fields


class HostGroupConstruct(core.Object):
    """
    L1 Construct: Nagios::Object::HostGroup
    """

    class Meta:
        object_type = 'hostgroup'

    hostgroup_name = fields.StringField(primary_key=True, required=True)
    alias = fields.StringField(required=True)
    members = fields.ForeignKey(relation='Host')
    hostgroup_members = fields.ForeignKey(relation='HostGroup')
    notes = fields.StringField()
    notes_url = fields.StringField()
    action_url = fields.StringField()

    def __init__(self, stack, hostgroup_name, alias, members=None,
                 hostgroup_members=None, notes=None, notes_url=None,
                 action_url=None):
        super().__init__(stack, hostgroup_name=hostgroup_name, alias=alias,
                         members=members, hostgroup_members=hostgroup_members,
                         notes=notes, notes_url=notes_url, action_url=action_url)


class HostGroup(HostGroupConstruct):
    """
    L2 Construct: Nagios::Object::HostGroup
    """

    def __init__(self, stack, hostgroup_name, alias, **kwargs):
        alias = alias or hostgroup_name
        super().__init__(
            stack, hostgroup_name=hostgroup_name, alias=alias, **kwargs)



import unittest

from ndk import core, objects
from ndk.objects import Command, CommandConstruct


class CommandTestCase(unittest.TestCase):

    def test_constructs_are_works(self):
        stack = core.Stack('CommandTesting')
        cmd = CommandConstruct(stack, command_name='top', command_line='top')
        assert cmd.command_name == 'top'

        cmd = Command(stack, command_name='ping',
                      command_line='ping 127.0.0.1')
        assert cmd.command_line == 'ping 127.0.0.1'

    def test_L3_construct_is_works(self):
        stack = core.Stack('CommandTesting')
        cmd = objects.command.Ping(stack, command_name='check-ping')
        assert cmd.command_name == 'check-ping'
        assert 'ping' in cmd.command_line



import unittest

from ndk import core, objects
from ndk.objects import HostGroup, HostGroupConstruct


class HostGroupTestCase(unittest.TestCase):

    def test_L1_construct_is_works(self):
        stack = core.Stack('HostGroupTesting')
        cg = HostGroupConstruct(
            stack, hostgroup_name='foo', alias='bar', members='baz')
        assert cg.hostgroup_name == 'foo'

    def test_L2_construct_is_works(self):
        stack = core.Stack('HostGroupTesting')
        cg = HostGroup(stack, hostgroup_name='foo',
                       alias='bar', members=['bar', 'baz'])
        assert cg.members == ['bar', 'baz']



from collections import defaultdict

from ndk.exceptions import DuplicateError, IntegrityError
from ndk.fields import Field


class Stack:
    """A NDK stack.
    """

    def __init__(self, name):
        """Initialize a NDK stack.

        Args:
            name (str): the application name.
        """
        self.name = name
        self.objects = defaultdict(dict)

    def __iter__(self):
        """Iter all objects.

        Yields:
            obj (NagiosObject): the next obj in the range of the all objects.
        """
        for obj in self.objects.values():
            yield from obj.values()

    def push(self, obj):
        """Push a new Nagios Object to this stack.

        Args:
            obj (NagiosObject): The base class of Nagios.
        """
        if obj.pk in self.objects[obj.__object_type__]:
            raise DuplicateError(
                f'{obj.pk} already exist in {obj.__object_type__} objects')
        self.objects[obj.__object_type__][obj.pk] = obj

    def synth(self):
        """Synthesizes the Nagios objects for this stack"""
        return '\n'.join((obj.synth() for obj in self))


class ObjectMeta(type):
    """A metacalss for Nagios Object."""

    def __new__(mcls, name, bases, attrs):
        parents = [b for b in bases if isinstance(b, ObjectMeta)]
        if not parents:
            # Make sure to initialize only the subclasses of Object
            return type.__new__(mcls, name, bases, attrs)

        # parent attrs
        object_type = None
        object_attr = {}
        for parent in parents:
            if hasattr(parent, '__object_type__'):
                object_type = parent.__object_type__
            if hasattr(parent, '__mappings__'):
                object_attr.update(**parent.__mappings__)

        # create __object_type__
        try:
            attrs['__object_type__'] = attrs['Meta'].object_type
        except KeyError:
            attrs['__object_type__'] = object_type

        # create __mappings__
        object_attr.update(**attrs)
        mappings = {k: v for k, v in object_attr.items()
                    if isinstance(v, Field)}
        attrs['__mappings__'] = mappings

        # create __primary_key__
        primary_key = [k for k, v in mappings.items() if v.primary_key]
        attrs['__primary_key__'] = primary_key

        # create __composite_key__
        composite_key = [k for k, v in mappings.items() if v.composite_key]
        attrs['__composite_key__'] = composite_key

        # clean
        for directive in attrs['__mappings__'].keys():
            attrs.pop(directive, None)

        # validation
        if not attrs['__object_type__']:
            raise IntegrityError(f'`__object_type__` not Found: {name}')
        if not attrs['__mappings__']:
            raise IntegrityError(f'`__mappings__` not Found: {name}')
        if not attrs['__primary_key__']:
            raise IntegrityError(f'`__primary_key__` not Found: {name}')

        return super().__new__(mcls, name, bases, attrs)


class Object(dict, metaclass=ObjectMeta):
    """A base class of Nagios Object."""

    def __init__(self, stack, *args, **kwargs):
        super().__init__(*args, **kwargs)
        assert isinstance(stack, Stack), f'Stack is invalid: {stack}'
        for key, val in kwargs.items():
            setattr(self, key, val)
        stack.push(self)

    def __getattr__(self, key):
        return self[key]

    def __setattr__(self, key, value):
        assert key in self.__mappings__, f'Field name is invalid: {key}'
        self[key] = value

    def __str__(self):
        return self.pk

    @property
    def pk(self):
        pks = (Field.normalize_name(self[key]) for key in self.__primary_key__)
        return "::".join(pks)

    def is_valid(self):
        for key, field in self.__mappings__.items():
            if field.required and not any([self.get(key), field.default]):
                raise IntegrityError(
                    f'{key} field is required in {self.__class__.__qualname__}')

    def render(self):
        self.is_valid()
        yield 'define %s {' % self.__object_type__
        for name, field in self.__mappings__.items():
            directives_value = self.get(name, None) or field.default
            if directives_value is not None:
                yield f'    {name}    {field.serializer(directives_value)}'
        yield '}'

    def synth(self):
        """Synthesizes the Nagios objects self"""
        return '\n'.join(self.render())



from ndk import core, fields


class ServiceGroupConstruct(core.Object):
    """
    L1 Construct: Nagios::Object::ServiceGroup
    """

    class Meta:
        object_type = 'servicegroup'

    servicegroup_name = fields.StringField(primary_key=True, required=True)
    alias = fields.StringField(required=True)
    members = fields.ForeignKey(relation='Service')
    servicegroup_members = fields.ForeignKey(relation='ServiceGroup')
    notes = fields.StringField()
    notes_url = fields.StringField()
    action_url = fields.StringField()

    def __init__(
            self, stack, servicegroup_name, alias, members=None,
            servicegroup_members=None, notes=None, notes_url=None,
            action_url=None):
        super().__init__(
            stack, servicegroup_name=servicegroup_name, alias=alias,
            members=members, servicegroup_members=servicegroup_members,
            notes=notes, notes_url=notes_url, action_url=action_url)


class ServiceGroup(ServiceGroupConstruct):
    """
    L2 Construct: Nagios::Object::ServiceGroup
    """

    def __init__(self, stack, servicegroup_name, alias=None, **kwargs):
        alias = alias or servicegroup_name
        super().__init__(
            stack, servicegroup_name=servicegroup_name, alias=alias, **kwargs)



import ipaddress
import unittest
from enum import Enum

from ndk import exceptions, fields


class Ipv4FieldTestCase(unittest.TestCase):

    def test_Ipv4Field_is_works(self):
        f = fields.Ipv4Field()
        assert f.primary_key == False
        assert f.required == False
        assert f.composite_key == False
        assert f.default == None

    def test_type_conversion_of_default_in_Ipv4Field(self):
        with self.assertRaises(ipaddress.AddressValueError):
            fields.Ipv4Field(default='foo')

        f = fields.Ipv4Field(default='127.0.0.1')
        assert isinstance(f.default, ipaddress.IPv4Address)
        assert '127.0.0.1' == str(f.default)
        assert 2130706433 == int(f.default)


class EnumOptions(Enum):
    FOO = 'f'
    BAR = 'b'


class Options:
    FOOBAR = 'fb'


class ChoiceFieldTestCase(unittest.TestCase):

    def test_ChoiceField_is_works(self):
        # `.choices` is required
        with self.assertRaises(exceptions.IntegrityError):
            fields.ChoiceField()

        with self.assertRaises(exceptions.IntegrityError):
            fields.ChoiceField(choices=Options)

        f = fields.ChoiceField(choices=EnumOptions)
        assert f.primary_key == False
        assert f.required == False
        assert f.composite_key == False
        assert f.default == None

    def test_type_conversion_of_default_in_ChoiceField(self):
        f = fields.ChoiceField(choices=EnumOptions,
                               default=EnumOptions.FOO)
        assert f.default == 'f'

        f = fields.ChoiceField(choices=EnumOptions,
                               default=[EnumOptions.FOO, EnumOptions.BAR])
        assert f.default == 'f,b'



import unittest

from ndk import core, objects
from ndk.objects import ContactGroup, ContactGroupConstruct


class ContactGroupTestCase(unittest.TestCase):

    def test_L1_constructs_are_works(self):
        stack = core.Stack('ContactGroupTesting')
        cg = ContactGroupConstruct(
            stack, contactgroup_name='foo', alias='bar', members='baz')
        assert cg.contactgroup_name == 'foo'

    def test_L2_constructs_are_works(self):
        stack = core.Stack('ContactGroupTesting')
        cg = ContactGroup(stack, contactgroup_name='foo',
                          alias='bar', members=['bar', 'baz'])
        assert cg.members == ['bar', 'baz']



import unittest

from ndk import core, objects
from ndk.objects import (BusinessDay, TimePeriod, TimePeriodConstruct,
                         TwentyFourSeven)


class TimePeriodTestCase(unittest.TestCase):

    def test_L1_construct_is_works(self):
        stack = core.Stack('TimePeriodTesting')
        tp = TimePeriodConstruct(
            stack, timeperiod_name='foo', alias='bar', sunday='00:00-24:00')
        assert tp.pk == 'foo'
        assert tp.alias == 'bar'
        assert tp.sunday == '00:00-24:00'

    def test_L2_construct_is_works(self):
        stack = core.Stack('TimePeriodTesting')
        tp = TimePeriod(
            stack, timeperiod_name='foo', alias='bar', monday='00:00-24:00')
        assert tp.sunday is None
        assert tp.monday == '00:00-24:00'

    def test_twenty_four_seven_is_works(self):
        stack = core.Stack('TimePeriodTesting')
        tp = TwentyFourSeven(stack)
        assert tp.pk == '24x7'
        assert tp.sunday == '00:00-24:00'
        assert all([tp.sunday, tp.monday, tp.tuesday, tp.wednesday,
                    tp.thursday, tp.friday, tp.saturday])

        tp = BusinessDay(stack)
        assert tp.pk == '8x5'
        assert tp.monday == '08:00-17:00'
        assert all([tp.monday, tp.tuesday, tp.wednesday,
                    tp.thursday, tp.friday])



import unittest

from ndk import core, exceptions, fields, objects


class ForeignKeyTestCase(unittest.TestCase):

    def test_foreign_key_is_works(self):
        class Host(core.Object):
            class Meta:
                object_type = 'host'
            host_name = fields.StringField(primary_key=True)
            parents = fields.ForeignKey(relation='Host')

        stack = core.Stack('ForeignKeyTesting')
        Host(stack, host_name='foo', parents='bar')



import unittest

from ndk import core, objects
from ndk.objects import Host, HostConstruct


class HostTestCase(unittest.TestCase):

    def test_L1_construct_is_works(self):
        stack = core.Stack('HostTesting')
        _24x7 = objects.TwentyFourSeven(stack)
        cmd = objects.command.Ping(stack)  # fake cmd
        host = HostConstruct(
            stack, host_name='foo', alias='bar', address='127.0.0.1',
            max_check_attempts=3, check_period=5, notification_interval=60,
            notification_period=_24x7, check_command=cmd)
        assert host.host_name == 'foo'
        assert host.max_check_attempts == 3
        assert host.address == '127.0.0.1'

    def test_L2_construct_is_works(self):
        stack = core.Stack('HostTesting')
        _24x7 = objects.TwentyFourSeven(stack)
        cmd = objects.command.Ping(stack)  # fake cmd
        host = Host(stack, host_name='foo', address='localhost',
                    notification_period=_24x7, check_period=_24x7,
                    check_command=cmd)
        assert host.max_check_attempts == 5
        assert host.address == 'localhost'



from enum import Enum

from ndk import core, fields


class WEEKDAYS(Enum):
    MONDAY = 'mon'
    TUESDAY = 'tue'
    WEDNESDAY = 'wed'
    THURSDAY = 'thu'
    FRIDAY = 'fri'


class WEEKEND(Enum):
    SUNDAY = 'sun'
    SATURDAY = 'sat'


class TimePeriodConstruct(core.Object):
    """
    L1 Construct: Nagios::Object::TimePeriod

    This construct correspond directly to Command defined by Nagios.
    """

    class Meta:
        object_type = 'timeperiod'

    timeperiod_name = fields.StringField(primary_key=True, required=True)
    alias = fields.StringField(required=True)
    sunday = fields.StringField()
    monday = fields.StringField()
    tuesday = fields.StringField()
    wednesday = fields.StringField()
    thursday = fields.StringField()
    friday = fields.StringField()
    saturday = fields.StringField()

    def __init__(self, stack, timeperiod_name, alias,
                 sunday=None, monday=None, tuesday=None, wednesday=None,
                 thursday=None, friday=None, saturday=None):
        super().__init__(stack=stack, timeperiod_name=timeperiod_name,
                         alias=alias, sunday=sunday, monday=monday,
                         tuesday=tuesday, wednesday=wednesday,
                         thursday=thursday, friday=friday, saturday=saturday)


class TimePeriod(TimePeriodConstruct):
    """
    L2 Construct: Nagios::Object::TimePeriod

    Command encapsulate L1 modules, it is developed to address specific use 
    cases and sensible defaults.
    """

    def __init__(self, stack, timeperiod_name, alias=None, **kwargs):
        """
        Define TimePeriod of Nagios Object

        Note:
        `.alias` is required
        If `.alias` is not set, it is the same as the timeperiod_name.
        """
        alias = alias or timeperiod_name
        super().__init__(stack=stack,
                         timeperiod_name=timeperiod_name,
                         alias=alias,
                         **kwargs)


class TwentyFourSeven(TimePeriod):
    """
    L3 Construct: Nagios::Object::TimePeriod

    L3 declare a resource to create particular use cases.
    """

    def __init__(self, stack, timeperiod_name='24x7',
                 alias='twenty-four seven', sunday='00:00-24:00',
                 monday='00:00-24:00', tuesday='00:00-24:00',
                 wednesday='00:00-24:00', thursday='00:00-24:00',
                 friday='00:00-24:00', saturday='00:00-24:00'):
        """
        Define 24x7 TimePeriod of Nagios Object

        TwentyFourSeven means Sunday through Saturday from 00:00 to 24:00.
        """
        super().__init__(stack, timeperiod_name=timeperiod_name, alias=alias,
                         sunday=sunday, monday=monday, tuesday=tuesday, wednesday=wednesday,
                         thursday=thursday, friday=friday, saturday=saturday)


class BusinessDay(TimePeriod):
    """
    L3 Construct: Nagios::Object::TimePeriod

    L3 declare a resource to create particular use cases.
    """

    def __init__(
            self, stack, timeperiod_name='8x5', alias='business day',
            monday='08:00-17:00', tuesday='08:00-17:00',
            wednesday='08:00-17:00', thursday='08:00-17:00',
            friday='08:00-17:00', **kwargs):
        """
        Define 8x5 TimePeriod of Nagios Object

        A business day means Monday through Friday from 8 a.m. to 5 p.m..
        """
        super().__init__(stack, timeperiod_name=timeperiod_name, alias=alias,
                         monday=monday, tuesday=tuesday, wednesday=wednesday,
                         thursday=thursday, friday=friday, **kwargs)



import unittest

from ndk import core, objects
from ndk.objects import Service, ServiceConstruct


class ServiceTestCase(unittest.TestCase):

    def test_L1_construct_is_works(self):
        stack = core.Stack('ServiceTesting')
        _24x7 = objects.TwentyFourSeven(stack)
        cmd = objects.command.Ping(stack)  # fake cmd
        service = ServiceConstruct(
            stack, service_description='foo', host_name='bar',
            check_command=cmd, check_interval=5, retry_interval=1,
            check_period=_24x7, notification_interval=60,
            notification_period=_24x7)
        assert service.service_description == 'foo'

    def test_L2_construct_is_works(self):
        stack = core.Stack('ServiceTesting')
        service = Service(
            stack, service_description='foo', host_name='foo',
            check_command='asdf', notification_period='24x7')
        assert service.max_check_attempts == 5
